from .common import coerce_matrix, expand_array, open_file
from .fortran import read_fortran_rectangle, read_fortran_square, to_fortran
from .inro import peek_mdf, read_emx, read_mdf, to_emx, to_mdf
from .nwp import (read_nwp_base_network, read_nwp_exatts_list,
                  read_nwp_link_attributes, read_nwp_node_attributes,
                  read_nwp_traffic_results,
                  read_nwp_traffic_results_at_countpost,
                  read_nwp_transit_line_attributes, read_nwp_transit_network,
                  read_nwp_transit_result_summary,
                  read_nwp_transit_segment_results,
                  read_nwp_transit_station_results, read_nwp_transit_vehicles)
from .omx import read_omx, to_omx



from __future__ import annotations

from os import PathLike
from typing import Dict, Iterable, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..general import sort_nicely
from ..matrices import fast_unstack

try:
    import openmatrix as omx
except ImportError:
    omx = None

MATRIX_TYPES = Union[pd.DataFrame, pd.Series, NDArray]


if omx is not None:
    def read_omx(src_fp: Union[str, PathLike], *, tables: Iterable[str] = None, mapping: str = None, tall: bool = False,
                 raw: bool = False, squeeze: bool = True) -> Union[MATRIX_TYPES, Dict[str, MATRIX_TYPES]]:
        """
        Reads Open Matrix (OMX) files. An OMX file can contain multiple matrices, so this function
        typically returns a Dict.

        Args:
            src_fp (str | PathLike): OMX file from which to read. Cannot be an open file handler.
            tables (Iterable[str], optional): List of matrices to read from the file. If None, all matrices will be
                read.
            mapping (str, optional): The zone number mapping to use, if known in advance.
            tall (bool, optional) : If True, matrices will be returned in 1D format. Otherwise, a 2D object is returned.
            raw (bool, optional): If True, matrices will be returned as raw Numpy arrays. Otherwise, Pandas objects are
                returned
            squeeze (bool, optional): If True, and the file contains exactly one matrix, return that matrix instead of a
                Dict.

        Returns:
            The matrix, or matrices contained in the OMX file.

        """
        omx_file = omx.open_file(str(src_fp), mode='r')
        try:
            table_names: List[str] = sort_nicely(omx_file.list_matrices()) if tables is None else list(tables)

            if not raw:
                if mapping is None:
                    rows, columns = omx_file.shape()
                    if rows != columns:
                        raise NotImplementedError('Handling of non-square matrices not implemented yet')
                    labels = pd.Index(range(rows))
                else:
                    zone_mapping: Dict[int, int] = omx_file.mapping(mapping)
                    labels = pd.Index(zone_mapping.keys())
                if tall:
                    labels = pd.MultiIndex.from_product([labels, labels], names=['o', 'd'])

            retval = {}
            for name in table_names:
                matrix = np.array(omx_file[name])
                if tall:
                    n = matrix.shape[0] * matrix.shape[1]
                    matrix.shape = n

                if not raw:
                    if tall:
                        matrix = pd.Series(matrix, index=labels, name=name)
                    else:
                        matrix = pd.DataFrame(matrix, index=labels, columns=labels)

                retval[name] = matrix

            if (len(retval) == 1) and squeeze:
                return retval[table_names[0]]
            return retval
        finally:
            omx_file.close()


    def to_omx(dst_fp: Union[str, PathLike], tables: Dict[str, MATRIX_TYPES], *, zone_index: pd.Index = None,
               title: str = '', descriptions: Dict[str, str] = None, attrs: Dict[str, Dict] = None,
               mapping_name: str = 'zone_numbers'):
        """Creates a new (or overwrites an old) OMX file with a collection of matrices.

        Args:
            dst_fp (str | PathLike): OMX to write.
            tables (Dict[str, pd.DataFrame | pd.Series | np.ndarray]: Collection of matrices to write. MUST be a dict,
                to permit the encoding of matrix metadata, and must contain the same types: all Numpy arrays, all
                Series, or all DataFrames. Checking is done to ensure that all items have the same shape and labels.
            zone_index: (pd.Index, optional): Override zone labels to use. Generally only useful if writing a dict of
                raw Numpy arrays.
            title (str, optional): The title saved in the OMX file.
            descriptions (Dict[str, str], optional): A dict of descriptions (one for each given matrix).
            attrs (Dict[str, Dict], optional): A dict of dicts (one for each given matrix).
            mapping_name (str, optional): Name of the mapping internal to the OMX file
        """

        matrices, zone_index = _prep_matrix_dict(tables, zone_index)

        if descriptions is None:
            descriptions = {name: '' for name in matrices.keys()}
        if attrs is None:
            attrs = {name: None for name in matrices.keys()}

        omx_file = omx.open_file(str(dst_fp), mode='w', title=title)
        try:
            omx_file.create_mapping(mapping_name, zone_index.tolist())
            for name, array in matrices.items():
                description = descriptions[name]
                attr = attrs[name]
                omx_file.create_matrix(name, obj=np.ascontiguousarray(array), title=description, attrs=attr)
        finally:
            omx_file.close()


    def _prep_matrix_dict(matrices: Dict[str, MATRIX_TYPES],
                          desired_zone_index: pd.Index) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        collection_type = _check_types(matrices)

        if collection_type == 'RAW':
            checked, n = _check_raw_matrices(matrices)
            zone_index = pd.Index(range(n))
        elif collection_type == 'SERIES':
            checked, zone_index = _check_matrix_series(matrices)
        elif collection_type == 'FRAME':
            checked, zone_index = _check_matrix_frames(matrices)
        else:
            raise NotImplementedError(collection_type)

        if desired_zone_index is not None:
            assert desired_zone_index.equals(zone_index)

        return checked, zone_index


    def _check_types(matrices: Dict[str, MATRIX_TYPES]) -> str:
        gen = iter(matrices.values())
        first = next(gen)

        item_type = 'RAW'
        if isinstance(first, pd.Series):
            item_type = 'SERIES'
        elif isinstance(first, pd.DataFrame):
            item_type = 'FRAME'

        msg = "All items must be the same type"

        for item in gen:
            if item_type == 'FRAME':
                assert isinstance(item, pd.DataFrame), msg
            elif item_type == 'SERIES':
                assert isinstance(item, pd.Series), msg
            else:
                assert isinstance(item, np.ndarray), msg

        return item_type


    def _check_raw_matrices(matrices: Dict[str, np.ndarray]) -> Tuple[Dict[str, np.ndarray], int]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        n_dim = len(matrix.shape)
        if n_dim == 1:
            shape = matrix.shape[0]
            n = int(shape ** 0.5)
            assert n * n == shape, "Only tall matrices that decompose to square shapes are permitted."
            matrix = matrix[...]
            matrix.shape = n, n
        elif n_dim == 2:
            n, cols = matrix.shape
            assert n == cols, "Only square matrices are permitted"
        else:
            raise TypeError("Only 1D and 2D arrays can be saved to OMX files")

        retval = {name: matrix}
        for name, matrix in gen:
            assert len(matrix.shape) == n_dim

            if n_dim == 1:
                assert matrix.shape[0] == (n * n)
                matrix = matrix[...]
                matrix.shape = n, n
            else:
                assert matrix.shape == (n, n)

            retval[name] = matrix

        return retval, n


    def _check_matrix_series(matrices: Dict[str, pd.Series]) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        tall_index = matrix.index
        assert tall_index.nlevels == 2
        matrix = matrix.unstack()
        zone_index = matrix.index
        assert zone_index.equals(matrix.columns)

        retval = {name: matrix.to_numpy()}
        for name, matrix in gen:
            assert tall_index.equals(matrix.index)
            matrix = fast_unstack(matrix, zone_index, zone_index).to_numpy()
            retval[name] = matrix

        return retval, zone_index


    def _check_matrix_frames(matrices: Dict[str, pd.DataFrame]) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        zone_index = matrix.index
        assert zone_index.equals(matrix.columns)

        retval = {name: matrix.to_numpy()}
        for name, matrix in gen:
            assert zone_index.equals(matrix.index)
            assert zone_index.equals(matrix.columns)
            retval[name] = matrix.to_numpy()

        return retval, zone_index
else:
    def read_omx(*args, **kwargs):
        raise NotImplementedError()

    def to_omx(*args, **kwargs):
        raise NotImplementedError()



from __future__ import annotations

from contextlib import contextmanager
from io import FileIO
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray


def coerce_matrix(matrix: Union[NDArray, pd.DataFrame, pd.Series], *, allow_raw: bool = True,
                  force_square: bool = True) -> NDArray:
    """Infers a NumPy array from given input

    Args:
        matrix (NDArray | pandas.DataFrame | pandas.Series):
        allow_raw (bool, optional): Defaults to ``True``.
        force_square (bool, optional): Defaults to ``True``.

    Returns:
        NDArray: A 2D ndarray of type float32
    """
    if isinstance(matrix, pd.DataFrame):
        if force_square:
            assert matrix.index.equals(matrix.columns)
        return matrix.values.astype(np.float32)
    elif isinstance(matrix, pd.Series):
        assert matrix.index.nlevels == 2, "Cannot infer a matrix from a Series with more or fewer than 2 levels"
        wide = matrix.unstack()

        union = wide.index | wide.columns
        wide = wide.reindex_axis(union, fill_value=0.0, axis=0).reindex_axis(union, fill_value=0.0, axis=1)
        return wide.values.astype(np.float32)

    if not allow_raw:
        raise NotImplementedError()

    matrix = np.array(matrix, dtype=np.float32)
    assert len(matrix.shape) == 2
    i, j = matrix.shape
    assert i == j

    return matrix


def expand_array(a: NDArray, n: NDArray, *, axis: int = None) -> NDArray:
    """Expands an array across all dimensions by a set amount

    Args:
        a (NDArray): The array to expand
        n (NDArray): The (non-negative) number of items to expand by.
        axis (int, optional): Defaults to ``None``. The axis to expand along, or None to expand along all axes.

    Returns:
        NDArray: The expanded array
    """

    if axis is None:
        new_shape = [dim + n for dim in a.shape]
    else:
        new_shape = []
        for i, dim in enumerate(a.shape):
            dim += n if i == axis else 0
            new_shape.append(dim)

    out = np.zeros(new_shape, dtype=a.dtype)

    indexer = [slice(0, dim) for dim in a.shape]
    out[indexer] = a

    return out


@contextmanager
def open_file(file_handle: Union[str, Path, FileIO], **kwargs):
    """Context manager for opening files provided as several different types. Supports a file handler as a str, unicode,
    ``pathlib.Path``, or an already-opened handler.

    Args:
        file_handle (str | Path | FileIO): The item to be opened or is already open.
        **kwargs: Keyword args passed to ``open()``. Usually mode='w'.

    Yields:
        File:
            The opened file handler. Automatically closed once out of context.
    """
    opened = False
    if isinstance(file_handle, str):
        f = open(file_handle, **kwargs)
        opened = True
    elif Path is not None and isinstance(file_handle, Path):
        f = file_handle.open(**kwargs)
        opened = True
    else:
        f = file_handle

    try:
        yield f
    finally:
        if opened:
            f.close()



from .general import (align_categories, is_identifier, reindex_series,
                      sort_nicely, sum_df_sequence)
from .io import *
from .matrices import (aggregate_matrix, disaggregate_matrix, fast_stack,
                       fast_unstack, matrix_balancing_1d, matrix_balancing_2d,
                       matrix_bucket_rounding, split_zone_in_matrix)
from .modelling import distance_array, distance_matrix, tlfd

try:
    from .plotting import (convergence_boxplot, location_summary,
                           trumpet_diagram)
except ImportError:
    pass
